#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2023-2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import unittest

import ibm_boto3
from ibm_watsonx_ai.experiment import AutoAI
from ibm_watsonx_ai.helpers.connections import DataConnection, S3Location
from ibm_watsonx_ai.wml_client_error import WMLClientError
from ibm_watsonx_ai.tests.utils import bucket_exists, create_bucket, is_cp4d
from ibm_watsonx_ai.tests.autoai.abstract_tests_classes import (
    AbstractTestAutoAIRemote)
from ibm_watsonx_ai.utils.autoai.errors import NoAutomatedHoldoutSplit


class TestAutoAIRemote(AbstractTestAutoAIRemote, unittest.TestCase):
    def test_00b_prepare_COS_instance(self):
        TestAutoAIRemote.cos_resource = ibm_boto3.resource(
            service_name="s3",
            endpoint_url=self.cos_endpoint,
            aws_access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
            aws_secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']
        )
        # Prepare bucket
        if not bucket_exists(self.cos_resource, self.bucket_name):
            TestAutoAIRemote.bucket_name = create_bucket(self.cos_resource, self.bucket_name)

            self.assertIsNotNone(TestAutoAIRemote.bucket_name)
            self.assertTrue(bucket_exists(self.cos_resource, TestAutoAIRemote.bucket_name))

        self.cos_resource.Bucket(self.bucket_name).upload_file(
            self.data_location,
            self.data_cos_path
        )

    def test_00c_prepare_connection_to_COS(self):
        self.api_client.set.default_project(self.project_id)
        connection_details = self.api_client.connections.create({
            'datasource_type': self.api_client.connections.get_datasource_type_uid_by_name('bluemixcloudobjectstorage'),
            'name': 'Connection to COS for tests',
            'properties': {
                'bucket': self.bucket_name,
                'access_key': self.cos_credentials['cos_hmac_keys']['access_key_id'],
                'secret_key': self.cos_credentials['cos_hmac_keys']['secret_access_key'],
                'iam_url': self.api_client.service_instance._href_definitions.get_iam_token_url(),
                'url': self.cos_endpoint
            }
        })

        TestAutoAIRemote.connection_id = self.api_client.connections.get_id(connection_details)
        self.assertIsInstance(self.connection_id, str)

    def test_00d_prepare_connected_data_asset(self):
        self.api_client.set.default_project(self.project_id)

        asset_jap_details = self.api_client.data_assets.store({
            self.api_client.data_assets.ConfigurationMetaNames.CONNECTION_ID: self.connection_id,
            self.api_client.data_assets.ConfigurationMetaNames.NAME: "Japanese",
            self.api_client.data_assets.ConfigurationMetaNames.DATA_CONTENT_NAME: '/' + self.bucket_name + '/' + self.data_cos_path
        })

        TestAutoAIRemote.asset_id = self.api_client.data_assets.get_id(asset_jap_details)
        self.assertIsInstance(self.asset_id, str)




    def test_00e_prepare_connected_data_asset_for_test_result_references(self):
        self.api_client.set.default_project(self.project_id)

        asset_details = self.api_client.data_assets.store({
            self.api_client.data_assets.ConfigurationMetaNames.CONNECTION_ID: self.connection_id,
            self.api_client.data_assets.ConfigurationMetaNames.NAME: "Iris - test asset",
            self.api_client.data_assets.ConfigurationMetaNames.DATA_CONTENT_NAME: '/' + self.bucket_name + '/' + self.data_cos_path
        })

        TestAutoAIRemote.test_asset_id = self.api_client.data_assets.get_id(asset_details)
        self.assertIsInstance(self.test_asset_id, str)

    def test_02_DataConnection_setup(self):
        TestAutoAIRemote.data_connection = DataConnection(data_asset_id=self.asset_id)
        TestAutoAIRemote.test_data_connection = DataConnection(data_asset_id=self.test_asset_id)
        TestAutoAIRemote.results_connection = DataConnection(
            connection_asset_id=self.connection_id,
            location=S3Location(
                bucket=self.bucket_name,
                path=self.results_cos_path
            )
        )

        self.assertIsNotNone(obj=TestAutoAIRemote.data_connection)
        self.assertIsNotNone(obj=TestAutoAIRemote.test_data_connection)
        self.assertIsNotNone(obj=TestAutoAIRemote.results_connection)

    def test_06b_get_test_data(self):  # test data
        test_df = self.remote_auto_pipelines.get_test_data_connections()[0].read()

        X_test = test_df.drop([self.prediction_column], axis=1).values
        y_test = test_df[self.prediction_column].values

        print("Test data sample:")
        print(X_test)
        print(y_test)

        self.assertGreater(len(X_test), 0)
        self.assertGreater(len(y_test), 0)

    def test_06c_get_train_data_assert(self):
        with self.assertRaises(NoAutomatedHoldoutSplit):
            X_train, X_holdout, y_train, y_holdout = self.remote_auto_pipelines.get_data_connections()[0].read(
                with_holdout_split=True)

    def test_06d_get_test_data_assert(self):
        with self.assertRaises(NoAutomatedHoldoutSplit):
            X_test, y_test = self.remote_auto_pipelines.get_test_data_connections()[0].read(with_holdout_split=True)


    def test_29_delete_connection_and_connected_data_asset(self):
        self.api_client.data_assets.delete(self.asset_id)
        self.api_client.connections.delete(self.connection_id)

        with self.assertRaises(WMLClientError):
            self.api_client.data_assets.get_details(self.asset_id)
            self.api_client.connections.get_details(self.connection_id)


if __name__ == '__main__':
    unittest.main()
